#!/usr/bin/perl
#
# Copyright (c) 1995 Wolfram Schneider <wosch@cs.tu-berlin.de>
# All rights reserved. Alle Rechte vorbehalten.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. All advertising materials mentioning features or use of this software
#    must display the following acknowledgement:
#    This product includes software developed by Wolfram Schneider
# 4. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# $Id: manck.pl,v 1.16 1995/08/12 22:28:10 wosch Exp $
#
# manck - check manual pages

sub var {
    $ENV{'PATH'} = '/bin:/usr/bin';
    $path = '/bin:/sbin:/usr/bin:/usr/sbin';
    $manpath = '/usr/share/man';
    $incpath = '/usr/include:/usr/X11/include:/usr/local/include';

    $ext = '.gz';
    $zcat = 'gzcat';
    $noman = 'No manual entry for'; # error message

    # 0: only errors 1: some more information 2: verbose
    $debug = 0;

    require "stat.pl";
    $| = 1;

    # bogus sections
    %s_bogus =
	(
	 '1-UCB', '1',
	 '3C',    '3',
	 '3C++',  '3',
	 '3N',    '3',
	 '3S',    '3',
	 '3X',    '3',
	 '3X11',  '3',
	 '3Xt',   '3',
	 '8C',    '8',
	 );

    # common sections
    %sh =
	(
	 'AUTHOR',            0,
	 'AUTHORS',           0,
	 'BUGS',              0,
	 'CAVEATS',           0,
	 'COMPATIBILITY',     0,
	 'DESCRIPTION',       0,
	 'DIAGNOSTICS',       0,
	 'ENVIRONMENT',       0,
	 'ERRORS',            0,
	 'EXAMPLE',           0,
	 'EXAMPLES',          0,
	 'FILES',             '&files',
	 'HISTORY',           0,
	 'NAME',              '&name',
	 'NOTES',             0,
	 'OPTIONS',           0,
	 'RETURN VALUE',      0,
	 'RETURN VALUES',     0,
	 'SEE ALSO',          '&see_also',
	 'STANDARDS',         0,
	 'SYNOPSIS',          '&synopsis',
	 );
}

sub usage {

    select STDERR;
    print <<EOF;
usage: manck [-d|-debug] [-d|-debug] [-h|-help] [-bin path]
             [-i|-include includepath] [-u=section]
             [-M mandir] [manpages ...]
EOF

    print "\nchecked sections: ";
    while(($key,$val) = each %sh) {
	print "'$key' " if $val;
    }
    print "\n";

    exit 1;
}

sub parse {
    local(*argv) = @_;

    while ($_ = $argv[0], /^-/) {
	shift @argv;
	last if /^--$/;
	if    (/^--?(d|debug)$/)        { $debug++ }
	elsif (/^--?(h|help|\?)$/)      { &usage }
	elsif (/^--?(b|bin)$/)          { exit(&bin($argv[0] || $path));  }
	elsif (/^--?M$/)	        { $manpath = $argv[0]; shift @argv;
					  $manpath =~ s=/+$==; }
	elsif (/^-i|-include$/)         { $incpath = $argv[0]; shift @argv; }
	elsif (/^-u=(.+)$/)	        { if (!defined $sh{$1}) {
	                                      warn "unknown section: $1\n";
                                              &usage;
					  }
					  $sh{$1} = 0 }
	else                            { &usage }
    }

    @include = split(/:/, $incpath);

    # use absolute path
    if ($manpath !~ "^/") {
	$manpath = ($ENV{'PWD'} || `pwd`) . '/' . $manpath;
	$manpath =~ s/\n//g;
    }

}

#
# check if binaries in $path have their own manpages
#
sub bin {
    local($path) = @_;
    local($dir);

    # common man section for unix commands
    #  '[168ln]([a-z]+)?';

    &cache_manpages(*man);

    foreach $dir (split(/:/, $path)) {
	print "$dir\n";
	opendir(DIR, $dir) || die "opendir: $dir $!\n";
	foreach (sort readdir(DIR)) {
	    next if $_ eq '.';
	    next if $_ eq '..';

	    if (! -e "$dir/$_" && -l "$dir/$_") {
		print "$_ is an empty symlink\n";
	    } elsif (! -x _ && -f _) {
		print "$_ is not executable for you\n";
	    }

	    if (!defined $man{$_}) {
		print "$noman $_\n";
	    } elsif ($man{$_} =~ /man\/man[168ln][a-z]*\/$_\.[168ln][a-z]*($ext)?(:|$)/) {		print "$_ in $man{$_}\n" if $debug > 0;	    }

	    else {
		print "Found $_ in $man{$_}\n";
	    }
	}
	closedir DIR;
    }
}

sub uniq_without_sort {
    local(@list) = @_;
    local(%found, $path, @list_new);

    while($path = shift @list) {
	if (!$found{$path}) {
	    push(@list_new, $path);
	}
	$found{$path} = 1;
    }

    @list_new;
}


#
# readdir all manpages from
# */man/man* directories
#
sub cache_manpages {
    local(*m) = @_;
    local($dir, $subdir, $file);
    local($mpath) = $manpath . ':' . $ENV{'MANPATH'};

    foreach $dir  (&uniq_without_sort(split(/:/, $mpath))) {
	opendir(DIR, $dir) || warn "$dir: $!\n";
	foreach $subdir (sort(readdir(DIR))) {
	    next if $subdir !~ /^man/;

	    opendir(SUBDIR, "$dir/$subdir") || warn "$dir/$subdir: $!\n";
	    foreach (readdir(SUBDIR)) {
		($file = $_) =~ s/$ext$//o;

		if ($file =~ /^(.+)\.[^\.]+$/) {
		    if (defined $m{$1}) {
			$m{$1} .= ':' . "$dir/$subdir/$_";
		    } else {
			$m{$1} .= "$dir/$subdir/$_";
		    }
		}
	    }
	    closedir SUBDIR;
	}
	closedir DIR;
    }
}

#
# like package &find()
#
sub read_manpages {
    local($mpath) = @_;
    local($dir, $subdir, $file);

    foreach $dir  (&uniq_without_sort(split(/:/, $mpath))) {
	opendir(DIR, $dir) || warn "$dir: $!\n";
	foreach $subdir (sort(readdir(DIR))) {
	    next if $subdir !~ /^man/;

	    opendir(SUBDIR, "$dir/$subdir") || warn "$dir/$subdir: $!\n";
	    foreach (sort readdir(SUBDIR)) {
		&manpage("$dir/$subdir/$_")
		    if /[^.]\.[^.]/;
	    }
	    closedir SUBDIR;
	}
	closedir DIR;
    }
}

#
# parse single manpage
# and analyze
#
sub manpage {
    local($file) = @_;
    local(@stat, $do, @list, $file2);

    unless ((@stat = stat($file))) {
	&err("$!\n");
	return 0;
    };

    if (! -f _) {
	&err("Not a plain file\n");
	return 0;
    }


    # cache
    if (defined $cached_file{$stat[$ST_DEV],$stat[$ST_INO]}) {
	&err("Already visit: " .
	    "$cached_file{$stat[$ST_DEV],$stat[$ST_INO]}\n") if $debug > 1;
	return 1;
    }
    $cached_file{$stat[$ST_DEV],$stat[$ST_INO]} = $file;

    if ($file =~ /$ext$/o) {
	$file2 = "$zcat $file |";
    } else {
	$file2 = $file;
    }
    open(MANPAGE, "$file2") || do {
	warn "open: $file2 $!\n";
	return 0;
    };

    $statistic{'manpage'}++;

    while(<MANPAGE>) {
	next if /^\.\\\"/o;	#" comment

	if (/^\.S[hH]\s+/o) {	# .SH or .Sh
	    # start action $do with following lines (until next .SH)
	    eval $do;

	    # reset action and array
	    $do = ''; @list = ();

	    $name = $_;
	    $name =~ s/^\.S[hH]\s+\"?//; #"
	    $name =~ s/\"?\s*$//;        #"

	    if ($sh{$name}) {
		&err("parse   $name\n") if $debug > 0;
		$do = $sh{$name}; # save action

	    } elsif (defined $sh{$name}) {
		&err("defined $name\n") if $debug > 1;
	    } else {
		&err("undef   $name\n") if $debug > 1;
	    }
	} elsif ($do) {
	    # \fBcurs_kernel\fR
	    s=\\f[BRPI]==g;
	    s=\\\&==g;

	    push(@list, $_);
	}
    }
    eval $do;			# last .SH in file
    close MANPAGE;
}

#
# locate manpage,
# may be from cache or with stat
#
sub locate {
    local($manpage, $section) = @_;


    if (defined $cached_mp{$manpage,$section}) {
	# failed from cache
	if ($cached_mp{$manpage,$section} < 0) {
	    $cached_mp{$manpage,$section}--;
	    &errS("$noman $manpage($section)\n");
	    return 0;
	}
	# found in cache
	else {
	    &err("Found $manpage($section)\n") if $debug > 1;
	    $cached_mp{$manpage,$section}++;
	    return 1;
	}
    }


    $cached_mp{$manpage,$section} = -1;

    # failed, cache is -1
    if (!defined $man{$manpage}) {
	&errS("$noman $manpage\n");
	return 0;
    }


    #
    # found manpage, cache = 1
    #

    # in specified section
    elsif ($man{$manpage} =~
	   /man\/man$section\/$manpage\.$section($ext)?(:|$)/) {
	&errS("Found $manpage($section) in $man{$manpage}\n")
	    if $debug > 1;
    }

    # in approximated section
    elsif (defined $s_bogus{$section} &&
	   $man{$manpage} =~ /man\/man$s_bogus{$section}\/$manpage\.$s_bogus{$section}($ext)?(:|$)/) {
	   &errS("Found $manpage($section) in section " .
	       "$s_bogus{$section}\n") if $debug > 0;

    # somewhere else
    } else {
	&errS("Found $manpage($section) in $man{$manpage}\n") if $debug > 0;
    }
    $cached_mp{$manpage,$section} = 1;
    return 1;
}


# print error messages
sub err {
    if (!$cache_err{$file}) {
	print "$file\n";
	$cache_err{$file} = 1;
    }
    print "\t";
    print @_;
}

# count errors 
sub errS {
    $statistic{$name}++;
    &err(@_);
}

#
# print statistic about references
# + successfully
# - failed
# 0 no references
#
sub references {
    local(@list) = ();
    local($key, $val);
    local(%noref);

    # + or -
    while(($key, $val) = each %cached_mp) {
	#$key =~ s/(.+)$/($1)/;
	$key =~ s/\034(.+)$/($1)/;
	push(@list, sprintf("%-4d %s", $val, $key));
    }

    # zero
    while(($key, $val) = each %cached_mp) {
	$key =~ s/\034.*$//;
	$noref{$key} = $val;
    }
    while(($key, $val) = each %cached_file) {
	# /usr/share/man/man1/(ls).1.gz
	if ($val =~ /\/([^\/]+)\.[^\.]+($ext)?$/o &&
	    !defined $noref{$1}) {
	    push(@list, sprintf("%-4d %s", 0, $val));
	}
    }


    print "\n" . "References count: successfully(+), failed(-), " .
	"whithout reference(0)\n" .
	join("\n",  reverse sort{$a <=> $b} @list) .
	    "\n";

    local($total) = $statistic{'manpage'};
    local($sum) = 0;
    delete $statistic{'manpage'};
    print "\nErrors in section\n";

    while(($key,$val) = each %statistic) {
	print sprintf("%-10s %5d\n", $key, $val);
	$sum += $val;
    }
    print "Manual pages absolut: $total errors absolut: $sum\n";
}


#
# .SH SEE ALSO
# test if manpages exist
#
sub see_also {
    foreach (@list) {
	s/\s+$//;		# strip ending blanks
	s/\"//g;		# remove '"'

	#if (/^$/) {
	#    &err("stop SEE ALSO empty line\n") if $debug > 1;
	#    return;
	#}

	foreach (split(/,/)) {
	    s/^\s+//;
	    #s=\\\&==;

	    # BSD
	    if (/^\.(Xr|Fn)\s+(\S+)\s+(\S+)/o) {
		&locate($2, $3); # && $debug > 1 && &err("found $2($3)\n");

	    }
	    # GNU
	    elsif (/^(\.[BI]R\s)?\s*(\S+)\s*\(\s*(\S+)\s*\)/o) {
		&locate($2, $3); # && $debug > 1 && &err("found $2($3)\n");
	    }
	    # Doku follow, break
	    elsif (/^\.(sp|Rs|LP|br|TP|RB|PP|Pp|%)/o ||
		   /^RFC/o) {
		&err("stop SEE ALSO: $_\n") if $debug > 1;
		return;
	    }

	    # Hm
	    elsif (/^\.(Bl|It)/) {
		&err("skip: `$_'\n") if $debug > 0;
	    }

	    # Garbage
	    else {
		if ($debug == 1) {
		    &err("unknown: `$_'\n");
		} elsif ($debug > 1) {
		    &errS("unknown: `$_'\n");
		}
	    }
	}
    }
}

#
# .SH FILES
# test if files exist
#
sub files {
    foreach (@list) {
	next if /^\.(Bl|El|\\\")/;

	# .TP \w'/etc/manpath.config'u+2n
	s/\s\\w'([^\']+)'.*/ $1/; #"

	if (s=.*\s(\S+)\s*$=$1= && /\//) {
	    s/\s+$//;
	    s/[\s\'\"].*//;	# "

	    next unless $_;

	    # only absolute path, no $HOME or ~/.foobar
	    next if $debug < 1 && !/^\//; 

	    &err("test -e $_\n") if $debug > 1;
	    if (! -e $_) {
		# spool/temp files
		if (/[\X\?*\]\#\$\[]$/ || /^[\~\$]/) {
		    if ($debug < 1) {
			return;
		    } elsif ($debug == 1) {
			&err("$_: file does not exist\n");
			return;
		    }
		}

		&errS("$_: file does not exist STAT\n");
	    }
	}
    }
}

#
# .SH SYNOPSIS
# test for include files
#
sub synopsis {
    local($inc);

    foreach (@list) {
	if (/\#include\s+[<"](\S+)[>"]/) { #
	    if ($cached_file{$1}) {
		&err("CACHED file: $1\n") if $debug > 2;
		next;
	    }

	    $cached_file{$1} = 1;

	    &errS("$1: include does not exist in: @include\n")
		unless &include_test($1);
	}
    }
}

sub include_test {
    foreach $inc (@include) {
	&err("test -f $inc/$1\n") if $debug > 1;
	return 1
	    if -f "$inc/$1";
    }
    return 0;
}

#
# .SH NAME
# test for NAME values
#
sub name {

    foreach (@list) {
	# BSD/doc style
	# .Nm name ...
	if (/^\.Nm\s+(.*)/) {
	    &name_test($1);
	} elsif (/^\.Nd/) {
	    return;		# ignore description
	}

	# GNU
	# name -   ...
	# name \-  ...
	elsif (/^\s*([a-zA-Z0-9].*)\s+(\\)?\-\s/) {
	    &name_test($1);
	}
	# name \(em ...
	# name \(mi ...
	elsif (/^\s*([a-zA-Z0-9_].*)\s+\\\((em|mi)/) {
	    &name_test($1);
	}
	# name, ...
	elsif (/^\s*([^,]+),\s+/) {
	    &name_test($1);
	}


	# rest, except empty lines
	elsif (!/^\s*$/) {
	    &err("$_");
	}

    }
}

sub name_test {
    local($page) = @_;

    foreach (split(/,/, $page)) {
	s/^\s+//; s/\s+$//;	# blanks
	#s/\\\&//;		# \&

	if ($man{$_}) {
	    &err("Found name $_\n") if $debug > 1;
	} else {
	    &errS("Manpage `$_' does not exist\n");
	}
    }
}

##
## Main
##
# read enviroment
&var;
&parse(*ARGV);

# single manpages
if ($#ARGV >= 0) {
    &cache_manpages(*man);
    foreach (@ARGV) { &manpage($_); }
}

# parse tree
else {
    &cache_manpages(*man);
    &read_manpages($manpath);
}

&references if $debug > 0;
